mod ancient_maintscript_entry;
mod ancient_python_version_field;
mod apache2_module_does_not_depend_on_apache2_api;
mod autotools_pkg_config_macro_not_cross_compilation_safe;
mod binary_control_field_duplicates_source;
mod binary_package_depends_on_toolchain_package;
mod bugs_field_does_not_refer_to_debian_infrastructure;
mod build_depends_on_1_revision;
mod build_depends_on_build_essential;
mod build_depends_on_obsolete_package;
mod build_depends_on_python_sphinx_only;
mod built_using_for_golang;
mod capitalization_error_in_description;
mod changelog_has_duplicate_line;
mod chown_with_dot;
mod circular_installation_prerequisite;
mod comma_separated_files_in_dep5_copyright;
mod common_license;
mod control_file_contains_dh_make_vcs_comment;
mod copyright_continued_lines_with_space;
mod copyright_format_uri;
mod copyright_has_crs;
mod copyright_missing_upstream_info;
mod copyright_refers_to_old_directory;
mod copyright_refers_to_symlink_license;
mod crlf_line_endings;
mod debcargo_collapse_features;
mod debhelper_but_no_misc_depends;
mod debhelper_compat_wrong_field;
mod debhelper_tools_from_autotools_dev_are_deprecated;
mod debian_changelog_file_contains_obsolete_user_emacs_settings;
mod debian_changelog_has_wrong_day_of_week;
mod debian_changelog_line_too_long;
mod debian_control_has_empty_field;
mod debian_control_has_obsolete_dbg_package;
mod debian_control_has_unusual_field_spacing;
mod debian_pycompat_is_obsolete;
mod debian_pyversions_is_obsolete;
mod debian_rules_contains_unnecessary_get_orig_source_target;
mod debian_rules_missing_recommended_target;
mod debian_rules_not_executable;
mod debian_rules_parses_dpkg_parsechangelog;
mod debian_rules_sets_dpkg_architecture_variable;
mod debian_rules_should_not_use_pwd;
mod debian_rules_uses_as_needed_linker_flag;
mod debian_rules_uses_deb_build_opts;
mod debian_rules_uses_deprecated_systemd_override;
mod debian_rules_uses_special_shell_variable;
mod debian_rules_uses_unnecessary_dh_argument;
mod debian_source_options_has_custom_compression_settings;
mod debian_tests_control_autodep8_is_obsolete;
mod debian_upstream_obsolete_path;
mod debian_watch_contains_dh_make_template;
mod debian_watch_file_is_missing;
mod debian_watch_file_old_format;
mod debian_watch_file_uses_deprecated_githubredir;
mod debian_watch_file_uses_github_releases;
mod debian_watch_file_uses_old_github_pattern;
mod debian_watch_not_mangling_version;
mod debian_watch_use_templates;
mod debian_watch_uses_insecure_uri;
mod debug_symbol_migration_possibly_complete;
mod declares_possibly_conflicting_debhelper_compat_versions;
mod default_mta_dependency_not_listed_first;
mod dep3_format_patch_author_or_from_is_better;
mod dep5_file_paragraph_references_header_paragraph;
mod depends_on_old_emacs;
mod depends_on_python_minimal;
mod description_contains_tabs;
mod description_synopsis_starts_with_article;
mod desktop_entry_contains_encoding_key;
mod desktop_entry_file_has_crs;
mod dh_clean_k_is_deprecated;
mod dh_installmanpages_is_obsolete;
mod dh_quilt_addon_but_quilt_source_format;
mod dm_upload_allowed;
mod dpatch_build_dep_but_no_patch_list;
mod duplicate_contact;
mod duplicate_override_context;
mod empty_debian_patches_series;
mod empty_debian_source_options;
mod empty_debian_tests_control;
mod essential_no_not_needed;
mod excessive_priority_for_library_package;
mod executable_desktop_file;
mod extended_description_contains_empty_paragraph;
mod field_name_typo_in_control;
mod field_name_typo_in_dep5_copyright;
mod field_name_typo_in_tests_control;
mod field_name_typo_in_upstream_metadata;
mod file_contains_trailing_whitespace;
mod font_package_not_multi_arch_foreign;
mod global_files_wildcard_not_first_paragraph_in_dep5_copyright;
mod homepage_field_uses_insecure_uri;
mod homepage_for_cpan_package_contains_version;
mod homepage_in_binary_package;
mod homepage_url_ends_with_dot_git;
mod initial_upload_closes_no_bugs;
mod invalid_dep3_format_patch_wrong_last_update;
mod invalid_short_name_in_dep5_copyright;
mod invalid_standards_version;
mod libmodule_build_perl_needs_to_be_in_build_depends;
mod libmodule_build_tiny_perl_needs_to_be_in_build_depends;
mod license_file_listed_in_debian_copyright;
mod maintainer_also_in_uploaders;
mod maintainer_script_empty;
mod maintainer_script_without_set_e;
mod malformed_override;
mod missing_build_dependency_for_dh_command;
mod missing_prerequisite_for_pyproject_backend;
mod missing_static_built_using_for_golang;
mod missing_vcs_browser_field;
mod needlessly_depends_on_awk;
mod new_package_uses_date_based_version_number;
mod newer_debconf_templates;
mod no_copyright_file;
mod no_homepage_field;
mod no_maintainer_field;
mod no_priority_field;
mod no_section_field;
mod obsolete_field_in_dep5_copyright;
mod obsolete_runtime_tests_restriction;
mod obsolete_url_in_packaging;
mod obsolete_vim_addon_manager;
mod old_fsf_address_in_copyright_file;
mod old_override_info_format;
mod orphaned_package_should_not_have_uploaders;
mod out_of_date_copyright_format_uri;
mod out_of_date_standards_version;
mod package_contains_linda_override;
mod package_has_no_description;
mod package_needs_versioned_debhelper_build_depends;
mod package_relation_with_perl_modules;
mod package_uses_deprecated_debhelper_compat_version;
mod package_uses_deprecated_dpatch_patch_system;
mod package_uses_deprecated_source_override_location;
mod patch_file_present_but_not_mentioned_in_series;
mod pkg_js_tools_test_is_missing;
mod pkg_perl_testsuite;
mod pkg_perl_vcs;
mod possible_missing_colon_in_closes;
mod priority_extra_is_replaced_by_priority_optional;
mod pubkey;
mod public_upstream_key_binary;
mod public_upstream_key_in_native_package;
mod public_upstream_key_not_minimal;
mod public_upstream_keys_in_multiple_locations;
mod python_teams_merged;
mod quilt_series_but_no_build_dep;
mod quilt_series_without_trailing_newline;
mod redundant_control_relation;
mod redundant_installation_prerequisite;
mod redundant_priority_optional_field;
mod renamed_tag;
#[doc = "Get the mapping of old lintian tag names to their current names."]
pub use renamed_tag::get_renamed_tags;
mod rules_not_executable;
mod rules_requires_root_missing;
mod skip_systemd_native_flag_missing_pre_depends;
mod source_format;
mod space_in_std_shortname_in_dep5_copyright;
mod systemd_service_alias_without_extension;
mod systemd_service_file_pidfile_refers_to_var_run;
mod systemd_service_file_refers_to_obsolete_bindto;
mod systemd_service_file_refers_to_obsolete_target;
mod systemd_service_file_shutdown_problems;
mod trailing_comma_in_maintainer_field;
mod transitional_package_should_be_oldlibs_optional;
mod typo_in_debhelper_override_target;
mod unnecessary_team_upload;
mod unnecessary_testsuite_autopkgtest_field;
mod unused_build_dependency_on_cdbs;
mod unused_entry_in_debian_source_include_binaries;
mod unused_license_paragraph_in_dep5_copyright;
mod unused_override;
mod unversioned_copyright_format_uri;
mod upstream_metadata_file;
mod upstream_metadata_has_obsolete_field;
mod upstream_metadata_in_native_source;
mod upstream_metadata_invalid;
mod useless_autoreconf_build_depends;
mod uses_debhelper_compat_file;
mod uses_deprecated_adttmp;
mod uses_dh_addons;
mod vcs_broken_uri;
mod vcs_field_bitrotted;
mod vcs_field_for_maintainer;
mod vcs_field_invalid_branch;
mod vcs_field_mismatch;
mod vcs_field_not_canonical;
mod vcs_field_uses_insecure_uri;
mod vcs_field_uses_not_recommended_uri_format;
mod vcs_git_uses_invalid_user_uri;
mod wrong_debian_qa_group_name;
mod wrong_section_according_to_package_name;
mod xc_package_type_in_debian_control;
mod xs_testsuite_field_in_debian_control;
mod xs_vcs_field_in_debian_control;
