#!/usr/bin/python3

from setuptools import setup

setup(
    name="dulwich",
    description="Python Git Library",
    version="0.1.1",
    author="Jelmer Vernooij",
    author_email="jelmer@jelmer.uk",
    url="https://www.dulwich.io/",
)
